# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from spike.common.message import Message

class TextMessage(Message):
    def __init__(self, msg_id, sender, epoch, text, reply_to, seen=False):
        super().__init__(msg_id, sender, epoch, reply_to, seen)
        self.text = text



# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from threading import Thread
from spike.tg.telegram import Telegram


class SeenMarker(Thread):

    def __init__(self, chat, messages):
        super().__init__()
        self.chat = chat
        self.messages = messages
        self.tg = Telegram().get_instance()


    def run(self):
        params = {
            "chat_id": self.chat.chat_id,
            "message_ids": [m.id for m in self.messages],
            "force_read": True
        }
        self.tg.call_method("viewMessages", params=params)

        for m in self.messages:
            m.seen = True
        self.chat.unreaded -= len(self.messages)
        if (self.chat.unreaded < 0):
            self.chat.unreaded = 0

        SeenMarker.chatlist.reload()



# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from sys import argv
from os import listdir, getpid
from os.path import basename


def spike_pid():
    cmd = basename(argv[0])
    current = str(getpid())
    proc_list = [x for x in listdir("/proc") if x.isdigit() and x != current]

    for proc in proc_list:
        with open("/proc/" + proc + "/cmdline") as f:
            cmdline = f.readline()
        if cmd in cmdline:
            return proc

    return None




# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import curses as cu
from spike.tg.telegram import Telegram
from spike.config.config import Config

class UserInput:
    KEY_SCROLL_UP_CHATLIST = 259   # Arrow up
    KEY_SCROLL_DOWN_CHATLIST = 258 # Arrow down
    KEY_ENTER_CHAT = 261           # Arrow right
    KEY_LEAVE_CHAT = 260           # Arrow left
    KEY_SCROLL_UP_DIALOG = 339     # Page up
    KEY_SCROLL_DOWN_DIALOG = 338   # Page down
    KEY_EXIT = 17                  # Ctrl + q
    KEY_BACKSPACE = (127, 263)
    KEY_ENTER = 10

    def __init__(self, stdscr, tui, textbox, chatlist, dialog, message_sender):
        self.stdscr = stdscr
        self.tui = tui
        self.textbox = textbox
        self.chatlist = chatlist
        self.dialog = dialog
        self.message_sender = message_sender


    def mainloop(self):
        while True:
            # wch can be str or int types. key ternary ensures int type
            wch = self.stdscr.get_wch()
            key = wch if isinstance(wch, int) else ord(wch)

            if key == UserInput.KEY_SCROLL_UP_CHATLIST:
                self.chatlist.scroll_up()
            elif key == UserInput.KEY_SCROLL_DOWN_CHATLIST:
                self.chatlist.scroll_down()
            elif key == UserInput.KEY_ENTER_CHAT:
                self.chatlist.enter_chat()
            elif key == UserInput.KEY_LEAVE_CHAT:
                self.chatlist.leave_chat()
            elif key == UserInput.KEY_SCROLL_UP_DIALOG:
                self.dialog.scroll_up()
            elif key == UserInput.KEY_SCROLL_DOWN_DIALOG:
                self.dialog.scroll_down()
            elif key in UserInput.KEY_BACKSPACE:
                self.textbox.delete_last_char()
            elif key == UserInput.KEY_EXIT:
                break
            elif key == UserInput.KEY_ENTER:
                self.message_sender.start()
            elif key == cu.KEY_RESIZE:
                self.tui.resize_windows()
            else:
                self.textbox.add_char(wch)

        if Config.get_instance().get_priv_online():
            Telegram().set_offline()



# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import curses as cu
from spike.config.config import Config
if Config.get_instance().get_emoji_convert_codes():
    import emoji

class TextBox:
    def __init__(self, window_size):
        lines, columns, start_y, start_x = window_size

        self.window = cu.newwin(lines, columns, start_y, start_x)
        self.window.scrollok(True)
        self.window.border()
        self.window.noutrefresh()
        cu.doupdate()

        self.cursor_x = self.cursor_y = 1
        # -2 because of borders
        self.writtable_cols = columns - 2
        self.writtable_lines = lines - 2
        self.scrolled_lines = 0
        self.text = []


    def _scroll_forward(self):
        self.window.scroll()
        self.window.move(self.cursor_y, 1)
        self.window.clrtoeol()  # Clear scrolled bottom border
        self.window.border()
        self.window.addch(0, self.writtable_cols+1, "↑")
        self.scrolled_lines += 1


    def _scroll_backwards(self):
        self.window.addstr(1, 1, "".join(self.text[-self.writtable_cols:]))
        self.cursor_y = 1
        self.cursor_x = self.writtable_cols
        self.scrolled_lines -= 1
        if self.scrolled_lines == 0:
            self.window.border()  # Delete arrow

    def add_char(self, ch):
        if self.cursor_x - 1>= self.writtable_cols:
            self.cursor_x = 1
            if self.cursor_y < self.writtable_lines:
                self.cursor_y += 1
            else:
                self._scroll_forward()
        self.window.addch(self.cursor_y, self.cursor_x, ch)
        self.cursor_x += 1
        self.window.noutrefresh()
        cu.doupdate()
        self.text.append(ch)


    def delete_last_char(self):
        # Do nothing if textbox is empty
        if self.scrolled_lines == 0 and self.cursor_x == 1 and self.cursor_y == 1:
            return
        # Jump to previous line if at the beggining of the line
        if self.cursor_x == 1:
            self.cursor_y -= 1
            self.cursor_x = self.writtable_cols + 1
        self.cursor_x -= 1
        self.window.addch(self.cursor_y, self.cursor_x, " ")
        self.text.pop()
        # Scroll backwards if there are hidden lines
        if self.scrolled_lines > 0 and self.cursor_x == 1 and self.cursor_y ==1:
            self._scroll_backwards()
            self.cursor_x += 1
        self.window.noutrefresh()
        cu.doupdate()


    def delete_text(self):
        self.window.clear()
        self.window.border()
        self.window.noutrefresh()
        cu.doupdate()
        self.cursor_x = self.cursor_y = 1
        self.scrolled_lines = 0
        self.text.clear()


    def get_text(self):
        if Config.get_instance().get_emoji_convert_codes():
            return emoji.emojize("".join(self.text))
        else:
            return "".join(self.text)



# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from threading import Thread
from spike.tg.telegram import Telegram


class MessageSender(Thread):

    def __init__(self, tui_textbox, tui_chatlist):
        super().__init__()
        self.tg = Telegram().get_instance()
        self.tui_textbox = tui_textbox
        self.tui_chatlist = tui_chatlist


    def run(self):
        if self.tui_chatlist.curr_chat_index > -1:
            chat = list(self.tui_chatlist.chats.keys())[self.tui_chatlist.curr_chat_index]
            chat_id = chat.chat_id

            text = self.tui_textbox.get_text()
            self.tui_textbox.delete_text()

            self.tg.send_message(chat_id=chat_id, text=text)
        super().__init__()  # To allow restarting the thread multiple times



# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import time
import curses as cu
from textwrap import wrap
from spike.tui import tui
from spike.common.textmessage import TextMessage
from spike.tg.seen_marker import SeenMarker
from spike.tg.chat_history import ChatHistory
from spike.config.config import Config

class Dialog:
    TEXT_MSG_PADDING = 2

    def __init__(self, window_size):
        lines, columns, start_y, start_x = window_size

        self.lines = lines
        self.available_width = columns - 2  # 2 of sides borders
        self.offset = 0
        self.reached_chat_start = False

        self.columns = columns
        self.window = cu.newwin(lines, columns, start_y, start_x)
        self.window.border()
        self.window.noutrefresh()
        cu.doupdate()


    def set_messages(self, chat, messages):
        self.offset = 0
        self.reached_chat_start = False
        self.messages = messages
        self.chat = chat


    def _display_text_message(self, msg):
        msg_lines = wrap(msg.text, self.available_width - \
                         Dialog.TEXT_MSG_PADDING)
        i = len(msg_lines) - 1
        while i >= 0 and self.available_lines > 0:
            # Check for empty lines
            if msg_lines[i][0] == "\n":
                self.available_lines -= 1
                msg_lines[i] = msg_lines[i].lstrip()


            self.window.addstr(self.available_lines, 1 +
                               Dialog.TEXT_MSG_PADDING, msg_lines[i])
            i -= 1
            self.available_lines -= 1


    def _display_header(self, msg):
        if self.available_lines > 0:
            sender = msg.sender[:self.available_width - 60]
            msg_time = time.strftime("%H:%M:%S", time.localtime(msg.epoch))
            msg_date = time.strftime("%d-%m-%Y", time.localtime(msg.epoch))
            id_hex = "ID:{}".format(format(msg.id, "X"))
            reply_to = "Reply:{}".format(format(msg.reply_to, "X"))

            offset = 1
            # Sender
            self.window.addstr(self.available_lines, offset, sender,
                                cu.color_pair(tui.Tui.COLOR_HEADER_USER))
            offset += len(sender) + 1
            # Time
            self.window.addstr(self.available_lines, offset, msg_time,
                                cu.color_pair(tui.Tui.COLOR_HEADER_TIME))
            offset += len(msg_time) + 1
            # Date
            self.window.addstr(self.available_lines, offset, msg_date,
                                cu.color_pair(tui.Tui.COLOR_HEADER_DATE))
            offset += len(msg_date) + 1
            # ID message
            self.window.addstr(self.available_lines, offset, id_hex,
                                cu.color_pair(tui.Tui.COLOR_HEADER_ID_MSG))
            offset += len(id_hex) + 1
            # Reply to
            if reply_to != "Reply:0":
                self.window.addstr(self.available_lines, offset, reply_to,
                                    cu.color_pair(tui.Tui.COLOR_HEADER_REPLY_TO))

            self.available_lines -= 1


    def _write_chat_name(self):
        name = "╼ " + self.chat.name[:self.columns - 6] + " ╾"
        half_len_name = int(len(name) / 2)
        half_columns = int(self.columns / 2)
        self.window.addstr(0, half_columns - half_len_name, name)


    def scroll_up(self):
        if self.offset < len(self.messages):
            self.offset += 1
            self.reload()
        elif not self.reached_chat_start:
            last_id = self.messages[0].id
            ChatHistory(self.chat.chat_id, self, last_id).retrieve()


    def scroll_down(self):
        if self.offset > 0:
            self.offset -= 1
            self.reload()


    def leave_dialog(self):
        self.offset = 0
        self.reached_chat_start = False
        self.messages = None
        self.chat = None

        self.window.clear()
        self.window.border()
        self.window.noutrefresh()
        cu.doupdate()


    def reload(self):
        self.window.clear()
        self.window.border()
        self._write_chat_name()

        if len(self.messages) == 0:
            if self.reached_chat_start:  # Clear and leave
                self.window.noutrefresh()
                cu.doupdate()
                return
            else:
                ChatHistory(self.chat.chat_id, self).retrieve()
        # If the number of preloaded messages is low and chat start not reached
        elif len(self.messages) < ChatHistory.NUMBER_MESSAGES_TO_RETRIEVE and \
                not self.reached_chat_start:
            ChatHistory(self.chat.chat_id, self, self.messages[0].id).retrieve()

        unreaded_messages = []
        self.available_lines = self.lines - 2  # 2: top and bottom borders
        i = len(self.messages) - 1 - self.offset
        while i >= 0 and self.available_lines > 0:
            msg = self.messages[i]

            # Check if is an unread message
            if not msg.seen:
                unreaded_messages.append(msg)

            # Classification of type of message
            if isinstance(msg, TextMessage):
                self._display_text_message(msg)
            elif self.available_lines > 0:
                self.window.addstr(self.available_lines, 1,
                                ">> Unhandled message type <<",
                                cu.color_pair(tui.Tui.COLOR_UNHANDLED_MSG) |
                                cu.A_BOLD)
                self.available_lines -= 1

            # Header printing
            self._display_header(msg)

            i -= 1
        self.window.noutrefresh()
        cu.doupdate()

        if Config.get_instance().get_priv_mark_as_seen():
            SeenMarker(self.chat, unreaded_messages).start()



# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from spike.common.textmessage import TextMessage
from spike.common.message import Message
from spike.tg.telegram import Telegram
from spike.config.config import Config
if Config.get_instance().get_emoji_convert_codes():
    import emoji

def create_message_object(msg_dict, users_dict):
    # Telegram instance
    tg = Telegram().get_instance()
    # Message ID fetch
    msg_id = msg_dict["id"]
    # Is message a reply?
    reply_to = msg_dict["reply_to_message_id"]
    # Message epoch fetch
    epoch = msg_dict["date"]
    # Message sender cached fetch
    if msg_dict["sender_user_id"] in list(users_dict.keys()):
        sender = users_dict[msg_dict["sender_user_id"]]
    else:  # Searches the username and caches it in self.user
        result = tg.get_user(msg_dict["sender_user_id"])
        result.wait()
        if result.error:
            sender = "<?>"
        else:
            users_dict[msg_dict["sender_user_id"]] = result.update["first_name"]
            sender = result.update["first_name"]

    # Message classification
    if msg_dict["content"]["@type"] == "messageText":
        if Config.get_instance().get_emoji_convert_codes():
            text = emoji.demojize(msg_dict["content"]["text"]["text"])
        else:
            text = msg_dict["content"]["text"]["text"]
        msg = TextMessage(msg_id, sender, epoch, text, reply_to)
    else:
        msg = Message(msg_id, sender, epoch, reply_to)

    return msg



# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import curses as cu
import locale
from sys import stderr
from spike.tui import chatlist
from spike.tui import dialog
from spike.tui import textbox
from spike.tui import userinput
from spike.common.chat import Chat  # For test
from spike.common.textmessage import TextMessage  # For test
from spike.tg.messages_updater import MessagesUpdater
from spike.tg.message_sender import MessageSender
from spike.tg.telegram import Telegram
from spike.tg.seen_marker import SeenMarker
from spike.tg.chat_history import ChatHistory
from spike.config.config import Config


class SmallTerminalError(Exception):
    def __init__(self, message):
        super().__init__(message)


class Tui:
    # Color pairs codes
    COLOR_SEL_CHAT = 1
    COLOR_DEFAULT = 2
    COLOR_UNHANDLED_MSG = 3
    COLOR_HEADER_USER = 4
    COLOR_HEADER_TIME = 5
    COLOR_HEADER_DATE = 6
    COLOR_HEADER_ID_MSG = 7
    COLOR_HEADER_REPLY_TO = 8
    # Windows proportions and sizes
    MINIMUM_WIDTH = 87
    MINIMUM_HEIGHT = 12
    TOP_LINES_PADDING = 1
    HEIGHT_TEXTBOX = 6
    CHATLIST_WIDTH_PROP = 0.25  # 1/4 of terminal width

    def __init__(self):
        # Set locale to user default in all categories
        locale.setlocale(locale.LC_ALL, "")
        try:
            cu.wrapper(self.start)
        except SmallTerminalError as ex:
            print(ex, file=stderr)


    def set_defaults(self):
        # Color pairs
        cu.use_default_colors()
        cu.init_pair(Tui.COLOR_SEL_CHAT, cu.COLOR_BLACK, cu.COLOR_BLUE)
        cu.init_pair(Tui.COLOR_DEFAULT, -1, -1)
        cu.init_pair(Tui.COLOR_UNHANDLED_MSG, cu.COLOR_WHITE, cu.COLOR_RED)
        cu.init_pair(Tui.COLOR_HEADER_USER, cu.COLOR_GREEN, -1)
        cu.init_pair(Tui.COLOR_HEADER_TIME, cu.COLOR_CYAN, -1)
        cu.init_pair(Tui.COLOR_HEADER_DATE, cu.COLOR_BLUE, -1)
        cu.init_pair(Tui.COLOR_HEADER_ID_MSG, cu.COLOR_RED, -1)
        cu.init_pair(Tui.COLOR_HEADER_REPLY_TO, cu.COLOR_YELLOW, -1)

        cu.curs_set(False)  # Hide cursor
        self.stdscr.attrset(cu.color_pair(Tui.COLOR_DEFAULT))
        self.stdscr.clear()
        width_chatlist = Tui.get_size_window_chatlist()[1]
        self.stdscr.addstr(0, int(width_chatlist/2) - 7, "Spike Messenger", cu.A_BOLD)
        self.stdscr.refresh()
        self.dialog = dialog.Dialog(Tui.get_size_window_dialog())
        self.chat_list = chatlist.ChatList(Tui.get_size_window_chatlist(),
                                          self.dialog)
        SeenMarker.chatlist = self.chat_list
        self.textbox = textbox.TextBox(Tui.get_size_window_textbox())
        self.message_sender = MessageSender(self.textbox, self.chat_list)
        self.userinput = userinput.UserInput(self.stdscr, self, self.textbox,
                            self.chat_list, self.dialog, self.message_sender)

        users_dict = dict()  # Cache of User ID and usernames
        ChatHistory.users_dict = users_dict
        self.messages_updater = MessagesUpdater(self.chat_list, self.dialog,
                                                users_dict)

    def _check_terminal_dimensions(self):
        if cu.COLS < Tui.MINIMUM_WIDTH:
            raise SmallTerminalError("Terminal width below minimum")
        if cu.LINES < Tui.MINIMUM_HEIGHT:
            raise SmallTerminalError("Terminal height below minimum")


    def start(self, stdscr):
        self.stdscr = stdscr
        self._check_terminal_dimensions()
        self.set_defaults()
        self.messages_updater.start()
        if Config.get_instance().get_priv_online():
            Telegram().set_online()
        self.userinput.mainloop()
        del(self.stdscr)  # So cu.wrapper can close everything


    def resize_windows(self):
        cu.update_lines_cols()
        self._check_terminal_dimensions()
        self.stdscr.clear()

        # Program header
        width_chatlist = Tui.get_size_window_chatlist()[1]
        self.stdscr.addstr(0, int(width_chatlist/2) - 7, "Spike Messenger", cu.A_BOLD)
        self.stdscr.noutrefresh()
        cu.doupdate()

        # Chatlist resize
        lines, columns, y, x = Tui.get_size_window_chatlist()
        self.chat_list.window.resize(lines, columns)
        self.chat_list.window.mvwin(y, x)
        self.chat_list.n_displayed_chats = lines - 4
        self.chat_list.writable_width = columns - 2
        self.chat_list.reload()

        # Dialog resize
        lines, columns, y, x = Tui.get_size_window_dialog()
        self.dialog.window.resize(lines, columns)
        self.dialog.window.mvwin(y, x)
        self.dialog.lines = lines
        self.dialog.columns = columns
        self.dialog.available_width = columns - 2
        if self.chat_list.curr_chat_index > -1:
            self.dialog.reload()
        else:
            self.dialog.leave_dialog()

        # Textbox resize
        lines, columns, y, x = Tui.get_size_window_textbox()
        self.textbox.window.resize(lines, columns)
        self.textbox.window.mvwin(y, x)
        self.textbox.writtable_cols = columns - 2
        self.textbox.writtable_lines = lines - 2
        text = self.textbox.get_text()
        self.textbox.delete_text()
        for c in text:
            self.textbox.add_char(c)


    def get_size_window_chatlist():
        lines = cu.LINES - Tui.TOP_LINES_PADDING
        columns = int(cu.COLS * Tui.CHATLIST_WIDTH_PROP)
        start_y = Tui.TOP_LINES_PADDING
        start_x = 0

        return (lines, columns, start_y, start_x)


    def get_size_window_dialog():
        chatlist_columns = Tui.get_size_window_chatlist()[1]
        lines = cu.LINES - Tui.TOP_LINES_PADDING - Tui.HEIGHT_TEXTBOX
        columns = cu.COLS - chatlist_columns - 1  # 1: Marging bewtween windows
        start_y = Tui.TOP_LINES_PADDING
        start_x = chatlist_columns + 1  # 1: Marging bewtween windows

        return (lines, columns, start_y, start_x)


    def get_size_window_textbox():
        dia_l, dia_c, dia_y, dia_x = Tui.get_size_window_dialog()
        lines = Tui.HEIGHT_TEXTBOX
        columns = dia_c
        start_y = Tui.TOP_LINES_PADDING + dia_l
        start_x = dia_x

        return (lines, columns, start_y, start_x)



# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from telegram.client import Telegram as T
from spike.config.config import Config

class Telegram:
    __instance = None

    def __init__(self):
        if Telegram.__instance == None:
            config = Config.get_instance()
            Telegram.__instance = T(
                api_id = config.get_tg_api_id(),
                api_hash = config.get_tg_api_hash(),
                phone = config.get_tg_phone(),
                database_encryption_key = config.get_tg_db_encryption_key(),
                files_directory = config.get_tg_files_directory()
            )
            self.__instance.login()

    @staticmethod
    def get_instance():
        if Telegram.__instance == None:
            Telegram()
        return Telegram.__instance


    def _set_status(self, status):
        params = {
            "name": "online",
            "value": status
        }
        self.__instance.call_method("setOption", params=params)


    def set_online(self):
        self._set_status(True)


    def set_offline(self):
        self._set_status(False)



# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from spike.tg.telegram import Telegram
from spike.tg.utils import create_message_object

class ChatHistory:

    NUMBER_MESSAGES_TO_RETRIEVE = 30

    def __init__(self, chat_id, tui_dialog, from_msg_id=0):
        self.chat_id = chat_id
        self.tui_dialog = tui_dialog
        self.from_msg_id = from_msg_id
        self.tg = Telegram().get_instance()


    def retrieve(self):
        left = ChatHistory.NUMBER_MESSAGES_TO_RETRIEVE
        raw_messages = []

        while left > 0:
            result = self.tg.get_chat_history(self.chat_id, left, self.from_msg_id)
            result.wait()

            if result.update["total_count"] == 0:
                self.tui_dialog.reached_chat_start = True
                break

            self.from_msg_id = result.update["messages"][-1]["id"]
            left -= result.update["total_count"]

            raw_messages.extend(result.update["messages"])

        for raw_msg in raw_messages:
            msg = create_message_object(raw_msg, ChatHistory.users_dict)
            self.tui_dialog.messages.insert(0, msg)



# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import curses as cu
from spike.tui import tui

class ChatList:
    def __init__(self, window_size, dialog):
        lines, columns, start_y, start_x = window_size

        self.first_disp_index = 0
        self.sel_chat_index = 0
        self.curr_chat_index = -1
        # -4 => two for margins, one for "Chats" banner and one for separator
        self.n_displayed_chats = lines - 4
        self.writable_width = columns - 2

        self.window = cu.newwin(lines, columns, start_y, start_x)
        self.window.border()

        self.dialog = dialog


    def set_chats(self, chats):
        self.chats = chats


    def scroll_down(self):
        # If at bottom
        if self.sel_chat_index == len(self.chats) - 1:
            return

        if self.sel_chat_index == self.first_disp_index + \
                                self.n_displayed_chats - 1:
            self.first_disp_index += 1
        self.sel_chat_index += 1
        self.reload()


    def scroll_up(self):
        # If at top
        if self.sel_chat_index == 0:
            return

        if self.sel_chat_index == self.first_disp_index:
            self.first_disp_index -= 1
        self.sel_chat_index -= 1
        self.reload()


    def enter_chat(self):
        self.curr_chat_index = self.sel_chat_index
        sel_chat = list(self.chats.keys())[self.sel_chat_index]
        self.dialog.set_messages(sel_chat, self.chats[sel_chat])
        self.dialog.reload()


    def leave_chat(self):
        self.curr_chat_index = -1
        self.dialog.leave_dialog()


    def reload(self):
        self.window.clear()
        self.window.border()
        # "Chats" banner
        self.window.addstr(1, 1, "Chats")
        self.window.addstr(2, 0, "├" + "─"*self.writable_width + "┤")

        i = self.first_disp_index  # index for self.chats
        j = 3  # index for self.window lines
        while i < len(self.chats) and i < (self.first_disp_index + \
                                self.n_displayed_chats):
            name = list(self.chats.keys())[i].name
            unreaded = list(self.chats.keys())[i].unreaded

            if unreaded > 0:
                text = "{} ({})".format(name[:self.writable_width - \
                                    (len(str(unreaded))+3)], unreaded)
            else:
                text = name[:self.writable_width]

            bold = cu.A_BOLD if unreaded else 0

            if i == self.sel_chat_index:
                self.window.addstr(j, 1, text,
                            cu.color_pair(tui.Tui.COLOR_SEL_CHAT) | bold)
            else:
                self.window.addstr(j, 1, text, bold)
            i += 1
            j += 1
        self.window.noutrefresh()
        cu.doupdate()



# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from configparser import ConfigParser
from os import environ, makedirs
from os.path import dirname, exists

class Config:
    __instance = None

    FILES_DIRECTORY = "{}/.local/share/spike-telegram".format(environ["HOME"])
    CONFIG_FILE = "{}/.config/spikerc".format(environ["HOME"])

    DEF_API_ID = "XXXXXXX"
    DEF_API_HASH = "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
    DEF_PHONE = "+YYXXXXXXXXX"
    DEF_DB_ENCRYPTION_KEY = ""
    DEF_NOTIFICATIONS = True
    DEF_SOUND = True
    DEF_ONLINE = True
    DEF_MARK_SEEN_MSG = True
    try:
        import emoji  # Check for the presence of emoji library
        DEF_EMOJI_CODES = True
    except ModuleNotFoundError:
        DEF_EMOJI_CODES = False

    def __init__(self):
        if Config.__instance == None:
            Config.__instance = self
            self.config_parser = ConfigParser()

            if exists(Config.CONFIG_FILE):
                self.config_parser.read(Config.CONFIG_FILE)
            else:
                self._write_default_config()
                print("Writing default config file to", Config.CONFIG_FILE)
                print("Please edit Telegram credentials and relaunch Spike")
                exit(0)


    @staticmethod
    def get_instance():
        if Config.__instance == None:
            Config()
        return Config.__instance


    def _write_config(self):
        if not exists(dirname(Config.CONFIG_FILE)):
            makedirs(dirname(Config.CONFIG_FILE))

        with open(Config.CONFIG_FILE, "w") as f:
            self.config_parser.write(f)


    def _write_default_config(self):
        self.config_parser["Telegram"] = {
            "api_id": Config.DEF_API_ID,
            "api_hash": Config.DEF_API_HASH,
            "phone": Config.DEF_PHONE,
            "db_encryption_key": Config.DEF_DB_ENCRYPTION_KEY,
        }
        self.config_parser["settings"] = {
            "notifications": Config.DEF_NOTIFICATIONS,
            "notifications_sound": Config.DEF_SOUND
        }
        self.config_parser["privacy"] = {
            "online_status": Config.DEF_ONLINE,
            "mark_as_seen": Config.DEF_MARK_SEEN_MSG
        }
        self.config_parser["emojis"] = {
            "convert_to_codes": Config.DEF_EMOJI_CODES
        }
        self._write_config()


    def get_tg_api_id(self):
        return self.config_parser["Telegram"].get("api_id", Config.DEF_API_ID)


    def get_tg_api_hash(self):
        return self.config_parser["Telegram"].get("api_hash", Config.DEF_API_HASH)


    def get_tg_phone(self):
        return self.config_parser["Telegram"].get("phone", Config.DEF_PHONE)


    def get_tg_db_encryption_key(self):
        return self.config_parser["Telegram"].get("db_encryption_key",
                                                  Config.DEF_DB_ENCRYPTION_KEY)

    def get_tg_files_directory(self):
        return Config.FILES_DIRECTORY

    def get_sttg_notifications(self):
        return self.config_parser["settings"].get("notifications",
                Config.DEF_NOTIFICATIONS).lower() == "true"

    def get_sttg_sound(self):
        return self.config_parser["settings"].get("notifications_sound",
                Config.DEF_SOUND).lower() == "true"

    def get_priv_online(self):
        return self.config_parser["privacy"].get("online_status",
                Config.DEF_ONLINE).lower() == "true"

    def get_priv_mark_as_seen(self):
        return self.config_parser["privacy"].get("mark_as_seen",
                Config.DEF_MARK_SEEN_MSG).lower() == "true"

    def get_emoji_convert_codes(self):
        return self.config_parser["emojis"].get("convert_to_codes",
                Config.DEF_EMOJI_CODES).lower() == "true"



# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from distutils.core import setup

setup(
    name="spike-telegram",
    version="0.1.0.1",
    packages=['spike', 'spike.tg', 'spike.common', 'spike.config', 'spike.tui'],
    license="GPL-3",
    author="Esteban López Rodríguez",
    author_email="gnu_stallman@protonmail.ch",
    url="https://gitlab.com/libre_hackerman/spike",
    description="Telegram client in Ncurses, written in Python",
    scripts=["bin/spike-telegram"],
    data_files=[("share/spike-telegram", ["resources/plup.wav"])],
    install_requires=["python-telegram", "emoji"]
)



# Copyright (C) 2020 Esteban López Rodríguez <gnu_stallman@protonmail.ch>
#
# This file is part of Spike.
#
# Spike is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Spike is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from threading import Thread
from _thread import start_new_thread
from os.path import join
from os import devnull
from sys import stderr, prefix
from subprocess import run
from spike.common import textmessage, message
from spike.common import chat as cht
from spike.tg.telegram import Telegram
from spike.tg.utils import create_message_object
from spike.config.config import Config


class MessagesUpdater(Thread):

    def __init__(self, tui_chatlist, tui_dialog, users):
        super().__init__()
        self.tg = Telegram().get_instance()
        self.tui_chatlist = tui_chatlist
        self.tui_dialog = tui_dialog
        self.users = users
        self.chats = dict()
        self.tui_chatlist.set_chats(self.chats)
        self.fnull= open(devnull, "w")


    def _get_own_id(self):
        result = self.tg.get_me()
        result.wait()
        self.own_id = result.update["id"]


    def _request_error(self, result):
        print("Error while calling Telegram API:", result.error_info)
        exit(1)


    def _retrieve_chatlists(self):
        # https://github.com/alexander-akhmetov/python-telegram/issues/58
        result = self.tg.get_chats(offset_order=9223372036854775807)
        result.wait()

        if result.error:
            self._request_error(result)

        chat_ids = result.update["chat_ids"]
        chats_modified = False
        for chat_id in chat_ids:
            # Add chat only if is not already in self.chats
            if not chat_id in [ch.chat_id for ch in self.chats.keys()]:
                chats_modified = True
                result = self.tg.get_chat(chat_id)
                result.wait()
                if result.error:
                    self._request_error(result)

                chat = cht.Chat(chat_id)
                if result.update["notification_settings"]["mute_for"] != 0:
                    chat.notifications = False
                chat.unreaded = result.update["unread_count"]
                if len(result.update["title"]) > 0:
                    chat.name = result.update["title"]
                else:
                    chat.name = "Cuenta eliminada"

                self.chats[chat] = []
        self.tui_chatlist.reload()


    def _message_handler(self, update):
        msg = create_message_object(update["message"], self.users)

        # Message chat fetch
        # Retrieve chats again if is not in current chat list
        if not update["message"]["chat_id"] in [ch.chat_id for ch in self.chats.keys()]:
            self._retrieve_chatlists()
        chat = None
        for ch in self.chats.keys():
            if ch.chat_id == update["message"]["chat_id"]:
                chat = ch
                break
        if not chat:
            print("Could not find chat for chat ID:", update["message"]["chat_id"],
                    file=stderr)
            exit(1)

        self.chats[chat].append(msg)
        chat.unreaded += 1
        if self.tui_chatlist.curr_chat_index > -1:
            selected_chat = list(self.chats.keys())[self.tui_chatlist.curr_chat_index]
        else:
            selected_chat = None

        # Reload dialog if the message is from selected chat
        if chat == selected_chat and \
                self.tui_dialog.offset == 0:  # Do not reload if scrolled
            self.tui_dialog.reload()

        # Reload chatlist to update unseen messages
        self.tui_chatlist.reload()

        # Notification
        if Config.get_instance().get_sttg_notifications() and \
                chat.notifications and \
                (chat != selected_chat or self.tui_dialog.offset > 0 ) and \
                update["message"]["sender_user_id"] != self.own_id:
            if isinstance(msg, textmessage.TextMessage):
                text = msg.text
            else:
                text = "New message"
            run(["notify-send", "-a", "Spike", chat.name, text])

            if Config.get_instance().get_sttg_sound():
                start_new_thread(run,
                        (["aplay", join(prefix, "share/spike-telegram/plup.wav")],),
                        {"stderr": self.fnull})

    def run(self):
        self._get_own_id()
        self._retrieve_chatlists()
        self.tg.add_message_handler(self._message_handler)
